;*=====================================================================*/
;*    .../prgm/project/bigloo/fthread/examples/recette/recette.scm     */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Mon Feb  4 14:28:58 2002                          */
;*    Last change :  Fri Mar  1 17:06:00 2002 (serrano)                */
;*    Copyright   :  2002 Manuel Serrano                               */
;*    -------------------------------------------------------------    */
;*    A test module that deploys the examples of SRFI18.               */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module recette
   (library fthread)
   (main    main))

;*---------------------------------------------------------------------*/
;*    err ...                                                          */
;*---------------------------------------------------------------------*/
(define (err . msg)
   (with-output-to-port (current-error-port)
      (lambda ()
	 (for-each write msg)
	 (newline))))

;*---------------------------------------------------------------------*/
;*    do-something-else ...                                            */
;*---------------------------------------------------------------------*/
(define (do-something-else)
   #unspecified)

;*---------------------------------------------------------------------*/
;*    *tests* ...                                                      */
;*---------------------------------------------------------------------*/
(define *tests* '())

;*---------------------------------------------------------------------*/
;*    test ...                                                         */
;*---------------------------------------------------------------------*/
(define (test name prgm::procedure res)
   (display* name "...")
   (default-scheduler (make-scheduler))
   (let ((provided (prgm)))
      (if (or (eq? res #unspecified)
	      (and (procedure? res) (res provided))
	      (equal? res provided))
	  (print "ok.")
	  (begin
	     (print "error.")
	     (print "   ==> provided: [" provided
		    "]\n       expected: ["
		    (if (procedure? res) (res 'result) res)
		    "]")))))

;*---------------------------------------------------------------------*/
;*    define-test ...                                                  */
;*---------------------------------------------------------------------*/
(define-macro (define-test id prgm . rest)
   (let ((t (match-case rest
	       ((:result ?result)
		`(list ',id (lambda () ,prgm) ,result))
	       (()
		`(list ',id (lambda () ,prgm) #unspecified))
	       (else
		(error "define-test" "Illegal rest argument" rest)))))
      `(set! *tests* (cons ,t *tests*))))

;*---------------------------------------------------------------------*/
;*    cond-expand ...                                                  */
;*---------------------------------------------------------------------*/
(define-test cond-expand
   (cond-expand (fthread #t)
		(else #f))
   :result #t)

;*---------------------------------------------------------------------*/
;*    cond-expand2 ...                                                 */
;*---------------------------------------------------------------------*/
(define-test cond-expand
   (eval '(cond-expand (fthread #t)
		       (else #f)))
   :result #t)

;*---------------------------------------------------------------------*/
;*    scheduler ...                                                    */
;*---------------------------------------------------------------------*/
(define-test scheduler
   (make-scheduler)
   :result scheduler?)

;*---------------------------------------------------------------------*/
;*    scheduler-react! ...                                             */
;*---------------------------------------------------------------------*/
(define-test scheduler-react
   (let ((s (make-scheduler)))
      (scheduler-react! s)
      (scheduler-react! s)
      #t)
   :result #t)

;*---------------------------------------------------------------------*/
;*    yield ...                                                        */
;*---------------------------------------------------------------------*/
(define-test yield
   (let ((res 0))
      (thread-start! (make-thread (lambda ()
				     (let loop ((i 22))
					(if (> i 0)
					    (begin
					       (set! res (+ res 1))
					       (thread-yield!)
					       (loop (- i 1))))))))
      (scheduler-start! 7)
      res)
   :result 7)

;*---------------------------------------------------------------------*/
;*    await ...                                                        */
;*---------------------------------------------------------------------*/
(define-test await
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-await! 'foo))))))
      (scheduler-start! 2)
      res)
   :result 0)

;*---------------------------------------------------------------------*/
;*    await2 ...                                                       */
;*---------------------------------------------------------------------*/
(define-test await2
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res))))
				 (thread-await! 'foo)))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 4))))))
      (scheduler-start! 2)
      res)
   :result 1)

;*---------------------------------------------------------------------*/
;*    await3 ...                                                       */
;*---------------------------------------------------------------------*/
(define-test await3
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ x res))))
				 (thread-await! 'foo)))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 1)
				 (broadcast 'bar 2)
				 (broadcast 'bar 3))))))
      (scheduler-start! 2)
      res)
   :result 6)

;*---------------------------------------------------------------------*/
;*    await4 ...                                                       */
;*---------------------------------------------------------------------*/
(define-test await4
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res))))
				 (thread-await! 'foo)))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 4))))))
      (scheduler-start! 2)
      res)
   :result 1)

;*---------------------------------------------------------------------*/
;*    await5 ...                                                       */
;*---------------------------------------------------------------------*/
(define-test await5
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res))))
				 (thread-await! 'foo)))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 4)
				 (thread-yield!)
				 (broadcast 'foo 4))))))
      (scheduler-start!)
      res)
   :result 1)

;*---------------------------------------------------------------------*/
;*    await6 ...                                                       */
;*---------------------------------------------------------------------*/
(define-test await6
   (let* ((res 0)
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 4)
				 (thread-yield!)
				 (broadcast 'foo 4)))))
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res))))
				 (thread-await! 'foo))))))
      (scheduler-start!)
      res)
   :result 1)

;*---------------------------------------------------------------------*/
;*    await7 ...                                                       */
;*---------------------------------------------------------------------*/
(define-test await7
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res))))
				 (thread-await! 'foo))
			      't1)))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 4)
				 (thread-yield!)
				 (broadcast 'foo 4))
			      't2))))
      (scheduler-start!)
      res)
   :result 1)

;*---------------------------------------------------------------------*/
;*    await8 ...                                                       */
;*---------------------------------------------------------------------*/
(define-test await8
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res))))
				 (thread-await! 'foo)
				 (thread-yield!))
			      't1)))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 4)
				 (thread-yield!)
				 (broadcast 'foo 4)
				 (thread-yield!))
			      't2))))
      (scheduler-start!)
      res)
   :result 1)

;*---------------------------------------------------------------------*/
;*    await9 ...                                                       */
;*---------------------------------------------------------------------*/
(define-test await9
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ x res)))))
			      't1)))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 1)
				 (broadcast 'bar 2)
				 (broadcast 'foo 2))
			      't2))))
      (scheduler-start!)
      res)
   :result 3)

;*---------------------------------------------------------------------*/
;*    await10 ...                                                      */
;*---------------------------------------------------------------------*/
(define-test await10
   (let* ((res 0)
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 4)
				 (broadcast 'foo 4)))))
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res)))) )))))
      (scheduler-start!)
      res)
   :result 1)

;*---------------------------------------------------------------------*/
;*    await11 ...                                                      */
;*---------------------------------------------------------------------*/
(define-test await11
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res))))))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'foo 4)
				 (broadcast 'bar 4))))))
      (scheduler-start!)
      res)
   :result 1)

;*---------------------------------------------------------------------*/
;*    await12 ...                                                      */
;*---------------------------------------------------------------------*/
(define-test await12
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res))))
				 (thread-yield!)
				 (thread-await! 'foo)))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 4)
				 (thread-sleep! 2)
				 (broadcast 'foo 4))))))
      (scheduler-start!)
      res)
   :result 1)

;*---------------------------------------------------------------------*/
;*    await13 ...                                                      */
;*---------------------------------------------------------------------*/
(define-test await13
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res))))
				 (thread-yield!)
				 (thread-await! 'foo)))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield!)
				 (broadcast 'foo 4)
				 (thread-yield!)
				 (broadcast 'bar 4)
				 (broadcast 'foo 4))))))
      (scheduler-start!)
      res)
   :result 0)

;*---------------------------------------------------------------------*/
;*    await14 ...                                                      */
;*---------------------------------------------------------------------*/
(define-test await14
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ x res))))
				 (thread-await! 'foo)
				 (set! res (+ 10 res))))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 1)
				 (broadcast 'bar 2)
				 (thread-yield!))))))

      (scheduler-start! 3)
      res)
   :result 3)

;*---------------------------------------------------------------------*/
;*    await-ntimes ...                                                 */
;*---------------------------------------------------------------------*/
(define-test await-ntimes
   (let ((t1 (thread-start! (make-thread
			     (lambda () (thread-await! 'foo 5))))))
      (scheduler-start!)
      (current-time))
   :result 7)

;*---------------------------------------------------------------------*/
;*    await-ntimes2 ...                                                */
;*---------------------------------------------------------------------*/
(define-test await-ntimes2
   (let ((t1 (thread-start! (make-thread
			     (lambda ()
				(thread-await! 'foo 5))
			     't1)))
	 (t2 (thread-start! (make-thread
			     (lambda ()
				(thread-yield!)
				(broadcast 'foo))
			     't2))))
      (scheduler-start!)
      (current-time))
   :result 2)

;*---------------------------------------------------------------------*/
;*    await-ntimes3 ...                                                */
;*---------------------------------------------------------------------*/
(define-test await-ntimes3
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-await! 'foo 5)
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res)))))))))
      (scheduler-start! 3)
      res)
   :result 0)

;*---------------------------------------------------------------------*/
;*    await-ntimes4 ...                                                */
;*---------------------------------------------------------------------*/
(define-test await-ntimes4
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-await! 'foo 5)
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res)))))
			      't1)))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 4))
			      't2))))
      (scheduler-start!)
      res)
   :result 0)

;*---------------------------------------------------------------------*/
;*    await-ntimes5 ...                                                */
;*---------------------------------------------------------------------*/
(define-test await-ntimes5
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res))))
				 (thread-await! 'foo 5)))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 4))))))
      (scheduler-start!)
      res)
   :result 1)

;*---------------------------------------------------------------------*/
;*    await-ntimes6 ...                                                */
;*---------------------------------------------------------------------*/
(define-test await-ntimes6
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res))))
				 (thread-await! 'foo 5)))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 4))))))
      (scheduler-start!)
      res)
   :result 1)

;*---------------------------------------------------------------------*/
;*    await-ntimes7 ...                                                */
;*---------------------------------------------------------------------*/
(define-test await-ntimes7
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield!)
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ x res))))
				 (thread-await! 'foo 5)))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield!)
				 (broadcast 'bar 1)
				 (broadcast 'bar 2)
				 (broadcast 'foo))))))
      (scheduler-start!)
      res)
   :result 3)

;*---------------------------------------------------------------------*/
;*    await-ntimes8 ...                                                */
;*---------------------------------------------------------------------*/
(define-test await-ntimes8
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ 1 res))))
				 (thread-await! 'foo 5)))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield!)
				 (broadcast 'foo)
				 (broadcast 'bar 4))))))
      (scheduler-start!)
      res)
   :result 0)

;*---------------------------------------------------------------------*/
;*    await-ntimes9 ...                                                */
;*---------------------------------------------------------------------*/
(define-test await-ntimes9
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-await! 'foo 5)
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ x res))))))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 4)
				 (thread-yield!)
				 (broadcast 'foo)
				 (broadcast 'bar 4))))))
      (scheduler-start!)
      res)
   :result 4)

;*---------------------------------------------------------------------*/
;*    await-ntimes10 ...                                               */
;*---------------------------------------------------------------------*/
(define-test await-ntimes10
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ x res))))
				 (thread-await! 'foo 5)))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 1)
				 (thread-yield!)
				 (broadcast 'bar 2)
				 (broadcast 'foo))))))
      (scheduler-start!)
      res)
   :result 1)

;*---------------------------------------------------------------------*/
;*    await-ntimes11 ...                                               */
;*---------------------------------------------------------------------*/
(define-test await-ntimes11
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ x res))))
				 (thread-await! 'foo 5)))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (broadcast 'bar 1)
				 (thread-yield!)
				 (broadcast 'bar 2))))))
      (scheduler-start!)
      (current-time))
   :result 8)

;*---------------------------------------------------------------------*/
;*    await-ntimes12 ...                                               */
;*---------------------------------------------------------------------*/
(define-test await-ntimes12
   (let* ((res 0)
	  (t1 (thread-start! (make-thread
			      (lambda ()
				 (thread-yield! :values 'bar
						(lambda (x)
						   (set! res (+ x res))))
				 (thread-await! 'foo 5)))))
	  (t2 (thread-start! (make-thread
			      (lambda ()
				 (thread-sleep! 10)
				 (broadcast 'bar 1)
				 (thread-yield!)
				 (broadcast 'bar 2))))))
      (scheduler-start!)
      res)
   :result 0)

;*---------------------------------------------------------------------*/
;*    wait-values ...                                                  */
;*---------------------------------------------------------------------*/
(define-test wait-values
   (let* ((res 0)
	  (mark #f)
	  (t (thread-start!
	      (make-thread
	       (lambda ()
		  (thread-yield! :values 'gee
				 (lambda (x)
				    (if (not mark)
					(begin
					   (set! mark #t)
					   (set! res (+ (current-time) res))))))
		  (set! mark #f)
		  (thread-yield! :values 'gee
				 (lambda (x)
				    (if (not mark)
					(begin
					   (set! mark #t)
					   (set! res (+ (current-time) res))))))
		  (set! mark #f)
		  (thread-yield! :values 'gee
				 (lambda (x)
				    (if (not mark)
					(begin
					   (set! mark #t)
					   (set! res (+ (current-time) res))))))))))
	  (t2 (thread-start!
	       (make-thread
		(lambda ()
		   (broadcast 'gee 1)
		   (broadcast 'gee 2)
		   (broadcast 'gee 3)
		   (thread-yield!)
		   (broadcast 'gee 1)
		   (broadcast 'gee 2)
		   (broadcast 'gee 3)
		   (thread-yield!)
		   (broadcast 'gee 1)
		   (broadcast 'gee 2)
		   (broadcast 'gee 3)
		   (thread-yield!)))))
	  (t3 (thread-start!
	       (make-thread
		(lambda ()
		   (broadcast 'gee 1)
		   (broadcast 'gee 2)
		   (broadcast 'gee 3)
		   (thread-yield!)
		   (broadcast 'gee 1)
		   (broadcast 'gee 2)
		   (broadcast 'gee 3)
		   (thread-yield!)
		   (broadcast 'gee 1)
		   (broadcast 'gee 2)
		   (broadcast 'gee 3)
		   (thread-yield!))))))
      (scheduler-start!)
      res)
   :result 6)

;*---------------------------------------------------------------------*/
;*    bind-exit ...                                                    */
;*---------------------------------------------------------------------*/
(define-test bind-exit
   (let ((res #f))
      (thread-start!
       (make-thread (lambda ()
		       (set! res
			     (bind-exit (stop)
				(+ 1 (stop #t)))))))
      (scheduler-start!)
      res)
   :result #t)

;*---------------------------------------------------------------------*/
;*    call/cc ...                                                      */
;*---------------------------------------------------------------------*/
(define-test call/cc
   (cond-expand
      (bigloo-c
       (let ((res #f))
	  (thread-start!
	   (make-thread (lambda ()
			   (set! res
				 (call/cc (lambda (kont)
					     (+ 14 (kont #t))))))))
	  (scheduler-start!)
	  res))
      (else
       #t))
   :result #t)

;*---------------------------------------------------------------------*/
;*    call/cc2                                                         */
;*---------------------------------------------------------------------*/
(define-test call/cc2
   (cond-expand
      (bigloo-c
       (let ((res 0))
	  (let ((t1 (thread-start!
		     (make-thread
		      (lambda ()
			 (set! res
			       (+ 1 (call/cc (lambda (kont)
						(broadcast 'kont kont)
						0)))))
		      't1)))
		(t2 (thread-start!
		     (make-thread
		      (lambda ()
			 (let ((kont (car (thread-await! 'kont))))
			    (try (kont 23)
				 (lambda (escape obj proc msg)
				    (set! res 36)
				    (escape #t)))))
		      't2))))
	     (scheduler-start!)
	     res)))
      (else
       36))
   :result 36)

;*---------------------------------------------------------------------*/
;*    main ...                                                         */
;*---------------------------------------------------------------------*/
(define (main argv)
   (let ((tests '()))
      (args-parse (cdr argv)
	 (("-h" "--help" (help "This help message"))
	  (args-parse-usage #f)
	  (exit 0))
	 (("-g" (help "Set debug mode"))
	  (set! *thread-debug* (+fx 1 *thread-debug*)))
	 (("-p" (help "Preserve thread execution order"))
	  (set! *thread-strict-order* #t))
	 (else
	  (set! tests (cons (string->symbol else) tests))))
      ;; run all the tests
      (for-each (lambda (pvn)
		   (apply test pvn))
		(if (null? tests)
		    (reverse *tests*)
		    (reverse (filter (lambda (t) (memq (car t) tests))
				     *tests*))))
      ;; if we reach that point, we are done
      (print "\n"
	     (if (null? tests) "All" (reverse tests))
	     " tests completed...")))


   

