;*=====================================================================*/
;*    serrano/prgm/project/bigloo/fthread/src/thread.scm               */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Mon Feb  4 11:49:11 2002                          */
;*    Last change :  Wed Jul 24 11:38:32 2002 (serrano)                */
;*    Copyright   :  2002 Manuel Serrano                               */
;*    -------------------------------------------------------------    */
;*    Bigloo Fair threads.                                             */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module __ft_thread

   (with   __ft_eval)
   
   (import __ft_types
	   __ft_%types
	   __ft_%thread
	   __ft_scheduler
	   __ft_%scheduler
	   __ft_env
	   __ft_%env
	   __ft_signal
	   __ft_reader
	   __ft_writer
	   __ft_socket
	   __ft_process
	   __ft_exception
	   __ft_%exception
	   __ft_%mutex)
   
   (static (class %thread::thread)

	   (class %thread-eq-sgn
	      (info::obj read-only))
	   (class %thread-term-sgn
	      (thread::thread read-only)))
   
   (export (make-thread::thread ::procedure . name)
	   (thread-sleep! ::int)
	   (thread-await! ::obj . timeout)
	   (thread-join! ::thread . timeout)
	   (current-thread)
	   (thread-start! ::thread . scheduler)
	   (thread-terminate! ::thread)
	   (thread-suspend! ::thread)
	   (thread-resume! ::thread)
	   (thread-yield! . arg)))

;*---------------------------------------------------------------------*/
;*    object-equal? ::%thread-term-sgn ...                             */
;*---------------------------------------------------------------------*/
(define-method (object-equal? o1::%thread-term-sgn o2)
   (and (%thread-term-sgn? o2)
	(eq? (%thread-term-sgn-thread o1)
	     (%thread-term-sgn-thread o2))))

;*---------------------------------------------------------------------*/
;*    object-equal? ::%thread-eq-sgn ...                               */
;*---------------------------------------------------------------------*/
(define-method (object-equal? o1::%thread-eq-sgn o2)
   (eq? o1 o2))

;*---------------------------------------------------------------------*/
;*    make-thread ...                                                  */
;*---------------------------------------------------------------------*/
(define (make-thread runnable . name)
   (define (thread-body t)
      (with-access::%thread t (%state %result %mutexes cleanup scheduler)
	 ;; terminate is used to abruptly terminate a thread
	 (let ((res (bind-exit (terminate)
		       (%thread-%terminate-set! t terminate)
		       (try (runnable)
			    (lambda (escape obj proc msg)
			       (notify-error obj proc msg)
			       (terminate #f))))))
	    ;; mark that the thread is now dead
	    (thread-set-dead! t)
	    ;; store the result of the thread
	    (set! %result res)
	    ;; abandon all the mutexes owned by the thread
	    (for-each abandon-mutex! %mutexes)
	    (set! %mutexes '())
	    ;; raise signal the thread termination
	    (broadcast (instantiate::%thread-term-sgn (thread t)) res)
	    ;; invoke the thread cleanup
	    (if (procedure? cleanup)
		(if (correct-arity? cleanup 1)
		    (cleanup res)
		    (error t "Illegal cleanup function" cleanup)))
	    ;; switch back to the scheduler
	    (thread-switch t scheduler))))
   (if (not (correct-arity? runnable 0))
       (error "make-thread"
	      "Illegal thread body (should be a procedure of 0 arguments)"
	      runnable)
       (letrec ((t (instantiate::%thread
		      (%builtin (%bgl-thread-new (lambda () (thread-body t))))
		      (name (if (pair? name) (car name) (gensym 'thread)))
		      (%exc-handlers (list uncaught-exception-hdl)))))
	  (thread-setup! t)
	  t)))

;*---------------------------------------------------------------------*/
;*    thread-sleep! ...                                                */
;*---------------------------------------------------------------------*/
(define (thread-sleep! timeout)
   (if (>fx timeout 0)
       (begin
	  (thread-yield!)
	  (thread-sleep! (-fx timeout 1)))))

;*---------------------------------------------------------------------*/
;*    thread-join! ...                                                 */
;*---------------------------------------------------------------------*/
(define (thread-join! t . args)
   (with-access::thread t (%state %result %exc-result %exc-raised)
      (if (or (eq? %state 'dying) (eq? %state 'dead))
	  (if %exc-raised
	      (raise %exc-result)
	      %result)
	  (match-case args
	     ((?to ?to-val)
	      (if (not (number? to))
		  (bigloo-type-error "thread-join" "integer" to)
		  (let ((v (thread-await! (instantiate::%thread-term-sgn
					     (thread t))
					  to)))
		     (cond
			((terminated-thread-exception? (thread-%exc-result t))
			 (raise (thread-%exc-result t)))
			((pair? v)
			 (car v))
			(else
			 to-val)))))
	     ((?to)
	      (if (not (number? to))
		  (bigloo-type-error "thread-join" "integer" to)
		  (let ((v (thread-await! (instantiate::%thread-term-sgn
					     (thread t))
					  to)))
		     (cond
			((terminated-thread-exception? (thread-%exc-result t))
			 (raise (thread-%exc-result t)))
			((pair? v)
			 (car v))
			(else
			 (raise *join-timeout-exception*))))))
	     (else
	      (let ((v (thread-await! (instantiate::%thread-term-sgn
					 (thread t)))))
		 (cond
		    ((terminated-thread-exception? (thread-%exc-result t))
		     (raise (thread-%exc-result t)))
		    (else
		     (car v)))))))))

;*---------------------------------------------------------------------*/
;*    thread-await! ...                                                */
;*---------------------------------------------------------------------*/
(define (thread-await! s . arg)
   (let* ((t (current-thread)))
      ;; await until the signal is present
      (define (await e s)
	 (let ((env (scheduler-env s)))
	    (let loop ()
	       (if (signal-lookup e env)
		   ;; the signal is present we return its value
		   (signal-values e env)
		   ;; the signal is absent, we block the thread
		   (let ((scdl (thread-scheduler t)))
		      (signal-register-thread! e env t)
		      (thread-set-await! t)
		      (thread-cooperate t)
		      (loop))))))
      ;; await at most n instants until the signal is present
      (define (await-ntimes e s timeout)
	 (if (not (integer? timeout))
	     (error "thread-await" "Illegal timeout" timeout)
	     (let ((env (scheduler-env s)))
		(let loop ((tmt timeout)
			   (sregp #f)
			   (eregp #f))
		   ;; mark that we have to do something at the end of instant
		   ;; for that thread
		   (cond
		      ((%scheduler-eoi s)
		       ;; we are at the end of instant
		       (thread-set-yield! t)
		       (thread-cooperate t)
		       (loop (- tmt 1) sregp #f))
		      ((signal-lookup e env)
		       ;; the signal is present we return its values
		       (if sregp (signal-unregister-thread! e env t))
		       (if eregp (thread-unset-eoi! t))
		       (signal-values e env))
		      ((< tmt 0)
		       ;; we have reached the timeout
		       (if sregp (signal-unregister-thread! e env t))
		       #f)
		      (else
		       ;; the signal is not present and we have not reached
		       ;; the end of instant
		       (if (not sregp) (signal-register-thread! e env t))
		       (if (not eregp) (thread-set-eoi! t))
		       (thread-set-await! t)
		       (thread-cooperate t)
		       (loop tmt #t #t)))))))
      (cond
	 ((not (thread? t))
	  (error "thread-await" "no such thread" t))
	 ((not (thread-attached? t))
	  (error "thread-await" "unattached thread" t))
	 (else
	  (with-access::thread t (scheduler)
	     (case s
		((:input input:)
		 (match-case arg
		    (((and ?p (? input-port?)) (and ?s (? integer?)) . ?tmt)
		     (let ((e (instantiate::%thread-eq-sgn
				 (info p))))
			;; run the io at the next instant
			(scheduler-exec scheduler (make-fair-reader p s e))
			;; block the thread
			(car (if (pair? tmt)
				 (await-ntimes e scheduler (car tmt))
				 (await e scheduler)))))
		    (((and ?p (? input-port?)) (and ?s (? string?)) . ?tmt)
		     (let ((e (instantiate::%thread-eq-sgn
				 (info p))))
			;; run the io at the next instant
			(scheduler-exec scheduler (make-fair-match p s e))
			;; block the thread
			(car (if (pair? tmt)
				 (await-ntimes e scheduler (car tmt))
				 (await e scheduler)))))
		    (((and ?p (? input-port?)) (and ?s (? pair?)) . ?tmt)
		     (if (pair? (filter (lambda (x) (not (char? x))) s))
			 (error "thread-await" "Illegal :input await" arg)
			 (let ((e (instantiate::%thread-eq-sgn
				     (info p))))
			    ;; run the io at the next instant
			    (scheduler-exec scheduler
					    (make-fair-memq p s e))
			    ;; block the thread
			    (car (if (pair? tmt)
				     (await-ntimes e scheduler (car tmt))
				     (await e scheduler))))))
		    (else
		     (error "thread-await" "Illegal :input await" arg))))
		((:output output:)
		 (match-case arg
		    (((and ?p (? output-port?)) (and ?s (? string?)) . ?tmt)
		     (let ((e (instantiate::%thread-eq-sgn
				 (info p))))
			;; run the io at the next instant
			(scheduler-exec scheduler (make-fair-writer p s e))
			;; block the thread
			(car (if (pair? tmt)
				 (await-ntimes e scheduler (car tmt))
				 (await e scheduler)))))
		    (else
		     (error "thread-await" "Illegal :output await" arg))))
		((:connect connect:)
		 (match-case arg
		    (((and ?s (? socket?)) . ?tmt)
		     (let ((e (instantiate::%thread-eq-sgn
				 (info s))))
			;; run the io at the next instant
			(scheduler-exec scheduler (make-fair-socket s e))
			;; block the thread
			(if (pair? tmt)
			    (await-ntimes e scheduler (car tmt))
			    (await e scheduler))))
		    (else
		     (error "thread-await" "Illegal :connect await" arg))))
		((:process process:)
		 (match-case arg
		    (((and ?p (? process?)) . ?tmt)
		     (let ((e (instantiate::%thread-eq-sgn
				 (info p))))
			;; run the io at the next instant
			(scheduler-exec scheduler (make-fair-process s e))
			;; block the thread
			(if (pair? tmt)
			    (await-ntimes e scheduler (car tmt))
			    (await e scheduler))))
		    (else
		     (error "thread-await" "Illegal :connect await" arg))))
		(else
		 (if (keyword? s)
		     (error "thread-await" "Illegal signal" s)
		     (if (and (pair? arg)
			      (number? (car arg))
			      (>= (car arg) 0))
			 (await-ntimes s scheduler (car arg))
			 (await s scheduler))))))))))

;*---------------------------------------------------------------------*/
;*    thread-yield! ...                                                */
;*---------------------------------------------------------------------*/
(define (thread-yield! . arg)
   (match-case arg
      ((:values ?e (and ?proc (? procedure?)))
       (if (correct-arity? proc 1)
	   (thread-wait-values! e proc)
	   (error "thread-yield!" "Illegal procedure yield arity" proc)))
      (else
       (let ((t (current-thread)))
	  (if (thread? t)
	      (with-access::thread t (scheduler)
		 ;; we now, really stop the thread
		 (thread-set-yield! t)
		 (thread-cooperate t)))))))

;*---------------------------------------------------------------------*/
;*    object-display ::thread ...                                      */
;*---------------------------------------------------------------------*/
(define-method (object-display o::thread . port)
   (with-output-to-port (if (pair? port) (car port) (current-output-port))
      (lambda ()
	 (with-access::thread o (name)
	    (display* "#<thread:" name ">")))))

;*---------------------------------------------------------------------*/
;*    object-write ::thread ...                                        */
;*---------------------------------------------------------------------*/
(define-method (object-write o::thread . port)
   (with-output-to-port (if (pair? port) (car port) (current-output-port))
      (lambda ()
	 (with-access::thread o (name %state)
	    (display* "#<thread:" name ":" %state ">")))))

;*---------------------------------------------------------------------*/
;*    object-print ::thread ...                                        */
;*---------------------------------------------------------------------*/
(define-method (object-print o::thread port print-slot)
   (object-write o port))

;*---------------------------------------------------------------------*/
;*    current-thread ...                                               */
;*---------------------------------------------------------------------*/
(define (current-thread)
   (if (not (scheduler? (current-scheduler)))
       #f
       (%scheduler-current-thread (current-scheduler))))

;*---------------------------------------------------------------------*/
;*    thread-start! ...                                                */
;*---------------------------------------------------------------------*/
(define (thread-start! t . s)
   ;; check that the thread is not already attached to a scheduler
   ;; (i.e. the thread is not already running in another scheduler)
   (if (thread-attached? t)
       (error "thread-start!" "thread already running" t)
       (let ((scdl (get-scheduler s 'thread-start!)))
	  ;; attach the thread to the scheduler
	  (with-access::thread t (scheduler %state)
	     (set! scheduler scdl)
	     ;; start the builtin thread
	     (%native-thread-start (thread-%builtin t))
	     (set! %state 'runnable))
	  ;; adding a thread simply append it to the list of thread
	  ;; to be started at the next scheduler instant
	  (with-access::%scheduler scdl (tostart threads startn)
	     ;; increment the number of started threads
	     (set! startn (+fx 1 startn))
	     (set! threads (cons t threads))
	     ;; thread are append in the inverse order.
	     (set! tostart (cons t tostart)))
	  ;; return the started thread
	  t)))
   
;*---------------------------------------------------------------------*/
;*    thread-terminate! ...                                            */
;*---------------------------------------------------------------------*/
(define (thread-terminate! t)
   (if (thread-attached? t)
       (if (not (memq (thread-%state t) '(dying dead)))
	   (with-access::thread t (scheduler %state %exc-result)
	      ;; mark the termination
	      (set! %exc-result *terminated-thread-exception*)
	      ;; adding a thread simply append it to the list of thread
	      ;; to be stoped at the next scheduler instant
	      (with-access::%scheduler scheduler (toterminate)
		 ;; thread are append in the inverse order.
		 (set! toterminate (cons t toterminate)))))))

;*---------------------------------------------------------------------*/
;*    thread-suspend! ...                                              */
;*---------------------------------------------------------------------*/
(define (thread-suspend! t)
   (if (thread-attached? t)
       (if (not (memq (thread-%state t) '(suspended dead)))
	   (let ((scdl (thread-scheduler t)))
	      ;; adding a thread simply append it to the list of thread
	      ;; to be stoped at the next scheduler instant
	      (with-access::%scheduler scdl (tosuspend)
		 ;; thread are append in the inverse order.
		 (set! tosuspend (cons t tosuspend)))))))

;*---------------------------------------------------------------------*/
;*    thread-resume! ...                                               */
;*---------------------------------------------------------------------*/
(define (thread-resume! t)
   (if (thread-attached? t)
       (if (eq? (thread-%state t) 'suspended)
	   (let ((scdl (thread-scheduler t)))
	      ;; adding a thread simply append it to the list of thread
	      ;; to be stoped at the next scheduler instant
	      (with-access::%scheduler scdl (toresume)
		 ;; thread are append in the inverse order.
		 (set! toresume (cons t toresume)))))))

;*---------------------------------------------------------------------*/
;*    thread-wait-values! ...                                          */
;*---------------------------------------------------------------------*/
(define (thread-wait-values! e thunk)
   (let ((t (current-thread)))
      (if (thread? t)
	  (let ((env (scheduler-env (thread-scheduler t))))
	     (thread-set-eoi! t)
	     ;; wait for the end of instant
	     (thread-cooperate t)
	     ;; run the event handler
	     (if (signal-lookup e env)
		 (let ((vals (signal-values e env)))
		    (if (pair? vals)
			(for-each (lambda (v) (thunk v)) vals)
			(if (correct-arity? thunk 0)
			    (t)
			    (t #unspecified)))))
	     (thread-set-yield! t)
	     (thread-cooperate t)))))
